from ase.db import connect
from ase.io import read, write
from ase.units import kB
from icet import (ClusterSpace, StructureContainer)
from icet.tools import enumerate_structures, map_structure_to_reference
from joblib import dump, load
import numpy as np
import sys, os

class setLattice():
  def __init__(self, primCell_path='../PRIM.POSCAR', cutoffs=[6., 6., 4.]):
    '''
    # the rest are only needed for 'asedb'
    primCell_path: primitive cell in POSCAR format with possible occupations after each site(atom position).
    cutoffs:       cutoffs for pairs, triplets, ... 
    '''

    # Read the primitive cell and possible chemical_symbols/occupations on each site
    self.primitive_structure, self.chemical_symbols = self.read_prim(primCell_path)
    # Set up the basic structure and a cluster space
    self.cs = ClusterSpace(structure = self.primitive_structure,
                           cutoffs   = cutoffs,
                           chemical_symbols = self.chemical_symbols, 
                           position_tolerance = 0.01,
                           )
    print(self.cs)

  def read_XY(self, data_path, ftype='asedb', Temp=1000):
    '''
    data_path:     path to training data file
    ftype:         training data type. 'asedb', structures and energies in ASE atom database; 
                                      'joblib', (correlation vectors, energies).
    Temp:          temperature used to calculate the partition function. A way to scale the Ytrain.
    '''

    X_train, Y_train = self.read_asedb(data_path=data_path, Temp=Temp)

    #if ftype == 'asedb' or data_path.endswith('.db'):
    #    X_train, Y_train = self.read_asedb(data_path=data_path, Temp=Temp)
    #elif ftype == 'joblib' or data_path.endswith('.joblib'):
    #    data = load(data_path) # format (X, Y)
    #    X_train = data[0]
    #    Y_train = data[1]
    return X_train, Y_train

  def read_prim(self, primCell_path):
    # Read the primitive cell structure (lattice)
    try: primitive_structure = read(primCell_path, format='vasp') # primitive structure
    except: 
        print("Failed to read the primitive cell as POSCAR:", primCell_path)
        sys.exit()

    # Adjusted from ASE read_vasp to get the checmical_symbols at each site
    fd = open(primCell_path, 'r')
    line1 = fd.readline()
    lattice_constant = float(fd.readline().split()[0])
    # Now the lattice vectors
    for ii in range(3):
        fd.readline()

    # Number of atoms. 
    atom_symbols = []
    numofatoms = fd.readline().split()
    # Check whether we have a VASP 4.x or 5.x format file. If the
    # format is 5.x, use the fifth line to provide information about
    # the atomic symbols.
    vasp5 = False
    try:
        int(numofatoms[0])
    except ValueError:
        vasp5 = True
        atomtypes = numofatoms
        numofatoms = fd.readline().split()

    # check for comments in numofatoms line and get rid of them if necessary
    commentcheck = np.array(['!' in s for s in numofatoms])
    if commentcheck.any():
        # only keep the elements up to the first including a '!':
        numofatoms = numofatoms[:np.arange(len(numofatoms))[commentcheck][0]]
    for i, num in enumerate(numofatoms):
        numofatoms[i] = int(num)

    # Check if Selective dynamics is switched on
    sdyn = fd.readline()
    selective_dynamics = sdyn[0].lower() == 's'

    # Check if atom coordinates are cartesian or direct
    if selective_dynamics:
        ac_type = fd.readline()
    else:
        ac_type = sdyn
    cartesian = ac_type[0].lower() == 'c' or ac_type[0].lower() == 'k'
    tot_natoms = sum(numofatoms)
    chemical_symbols = []
    for atom in range(tot_natoms):
        ac = fd.readline().split()
        occupations = ac[3:]
        chemical_symbols.append(occupations)
    print(chemical_symbols)

    return primitive_structure, chemical_symbols

  def read_asedb(self, data_path, Temp):
    '''
    data_path:     path to training data file
    Temp:          temperature used to calculate the partition function. A way to scale the Ytrain.
    cs:            cluster space which defines clusters to consider on a given lattice 
    '''

    # Load structures and energies from a ASE database
    try: 
        db = connect(data_path)
    except: 
        print("This data file cannot be loaded as ASE database:", data_path)
        sys.exit()

    sc = StructureContainer(cluster_space = self.cs)

    for row in db.select():
        structure = row.toatoms()
        try: energy = row.energy_perPd
        except: energy = None
        # --------------------------
        # This section only works for HxPd
        #coni = structure.get_chemical_symbols().count('H') /structure.get_chemical_symbols().count('Pd')
        #E0 = -5.220967 #Pd
        #E1 = -8.810319 #HPd_tetra
        #Eform = energy - coni*E1 - (1-coni)*E0
        #E0 = -16.842583 #TiFe
        #E2 = -23.860611 #H2TiFe
        #coni = structure.get_chemical_symbols().count('H') / (len(structure)/8.0)
        #Eform = row.energy_perPd - coni*0.5*E2 - (1-coni*0.5)*E0
        # --------------------------
        #partF = np.exp(-Eform/kB/Temp)
        #partF = -Eform
        partF = row.energy_perPd
        targetp = partF
        sc.add_structure(structure=row.toatoms(),
                         user_tag=row.tag,
                         properties={'targetp': targetp})

    tmpdata = sc.get_fit_data(key='targetp')
    X_train = tmpdata[0]
    Y_train = tmpdata[1]

    return X_train, Y_train

  def gen_struc_corr(self, struc_path, corr_path):
    '''
    function to generate new correlation vectors for sampling 
    corr_path:     path to existing corrlation vectors stored as a list in joblib format
    '''
    struc_pool = []
    if os.path.exists(struc_path): 
        #struc_pool = load(struc_path)
        struc_db = connect(struc_path)
        for row in struc_db.select():
            structure = row.toatoms()
            struc_pool.append(structure)
    else:
        print("No basis structure provided")
        print("Read CONTCARs from basis_path")
        infile = struc_path.split('.')[0]+'_path'
        with open(infile, 'r') as f:
           lines = f.readlines()
        struc_db = connect(struc_path)
        for pathi in lines:
           posi = os.path.join(pathi.strip('\n'), 'CONTCAR.vasp')
           atoms = read(posi, format='vasp')
           struc_pool.append(atoms)
           basis_db.write(atoms, path = pathi)

    if os.path.exists(corr_path): 
        corr_pool = load(corr_path)
    else:
        corr_pool = []
        reference = read('PRIM.POSCAR', format='vasp')
        for structure in struc_pool:
            print(structure)
            mapped_structure, info = map_structure_to_reference(structure, reference, inert_species=['O'])
            write('test.vasp', mapped_structure, format='vasp', vasp5=True, direct=True)
            corri = self.cs.get_cluster_vector(mapped_structure)
            corr_pool.append(corri)
        dump(corr_pool, corr_path)
    print('Number of structures as basis set:', len(struc_pool))
    print('Number of correlation vectors generatedd:', len(corr_pool))

    # output the concentration of each structure
    '''
    conc_pool = []
    for structure in struc_pool:
        conc = structure.get_chemical_symbols().count('H') / (len(structure)/8.0)
        conc_pool.append(conc)
    return struc_pool, corr_pool, conc_pool
    '''
    return struc_pool, corr_pool
        
